Index: user/test/com/google/gwt/user/server/rpc/RPCTest.java
===================================================================
--- user/test/com/google/gwt/user/server/rpc/RPCTest.java	(revision 0)
+++ user/test/com/google/gwt/user/server/rpc/RPCTest.java	(revision 0)
@@ -0,0 +1,394 @@
+package com.google.gwt.user.server.rpc;
+
+import com.google.gwt.user.client.rpc.RemoteService;
+import com.google.gwt.user.client.rpc.SerializableException;
+import com.google.gwt.user.client.rpc.SerializationException;
+
+import junit.framework.TestCase;
+
+import java.lang.reflect.Method;
+
+/**
+ * Tests for the {@link com.google.gwt.user.server.rpc.RPC RPC} class. 
+ */
+public class RPCTest extends TestCase {
+
+  private static interface A extends RemoteService {
+    void method1() throws SerializableException;
+    int method2();
+    int method3(int val);
+  }
+
+  private static interface B {
+    void method1();
+  }
+
+  private final String VALID_ENCODED_REQUEST = "0\uffff" + // version
+      "0\uffff" + // flags
+      "2\uffff" + // string table entry count
+      A.class.getName() + "\uffff" + // string table entry #0
+      "method2" + "\uffff" + // string table entry #1
+      "1\uffff" + // interface name
+      "2\uffff" + // method name
+      "0\uffff"; // param count
+
+  private final String INVALID_METHOD_REQUEST = "0\uffff" + // version
+  "0\uffff" + // flags
+  "2\uffff" + // string table entry count
+  A.class.getName() + "\uffff" + // string table entry #0
+  "method3" + "\uffff" + // string table entry #1
+  "1\uffff" + // interface name
+  "2\uffff" + // method name
+  "0\uffff"; // param count
+  
+  private final String INVALID_INTERFACE_REQUEST = "0\uffff" + // version
+      "0\uffff" + // flags
+      "2\uffff" + // string table entry count
+      B.class.getName() + "\uffff" + // string table entry #0
+      "method1" + "\uffff" + // string table entry #1
+      "1\uffff" + // interface name
+      "2\uffff" + // method name
+      "0\uffff"; // param count
+
+  /**
+   * Tests for method RPC.decodeRequest(String)
+   * 
+   * <p/> Cases:
+   * <ol>
+   * <li>String == null</li>
+   * <li>String == ""</li>
+   * <li>Valid request
+   * </ol>
+   */
+  public void testDecodeRequestString() {
+    // Case 1
+    try {
+      RPC.decodeRequest(null);
+      fail("Expected NullPointerException");
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (SerializationException e) {
+      fail(e.getMessage());
+    }
+
+    // Case 2
+    try {
+      RPC.decodeRequest("");
+      fail("Expected IllegalArgumentException");
+    } catch (IllegalArgumentException e) {
+      // expected to get here
+    } catch (SerializationException e) {
+      fail(e.getMessage());
+    }
+    
+    // Case 3
+    try {
+      RPCRequest request = RPC.decodeRequest(VALID_ENCODED_REQUEST);
+    } catch (SerializationException e) {
+      e.printStackTrace();
+    }
+  }
+
+  /**
+   * Tests for method RPC.decodeRequest(String, Class)
+   * 
+   * <p/> Cases:
+   * <ol>
+   * <li>String == null</li>
+   * <li>String == ""</li>
+   * <li>Class is null</li>
+   * <li>Class implements RemoteService subinterface</li>
+   * <li>Class implements the requested interface but it is not a subtype of
+   * RemoteService</li>
+   * <li>Class implements RemoteService derived interface but the method does
+   * not exist
+   * </ol>
+   * 
+   * @throws NoSuchMethodException
+   * @throws SecurityException
+   */
+  public void testDecodeRequestStringClass() throws SecurityException,
+      NoSuchMethodException {
+    // Case 1
+    try {
+      RPC.decodeRequest(null, null);
+      fail("Expected NullPointerException");
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (SerializationException e) {
+      fail(e.getMessage());
+    }
+
+    // Case 2
+    try {
+      RPC.decodeRequest("", null);
+      fail("Expected IllegalArgumentException");
+    } catch (IllegalArgumentException e) {
+      // expected to get here
+    } catch (SerializationException e) {
+      fail(e.getMessage());
+    }
+
+    // Case 3
+    try {
+      RPCRequest request = RPC.decodeRequest(VALID_ENCODED_REQUEST, null);
+      assertEquals(A.class.getMethod("method2", null), request.getMethod());
+      assertTrue(request.getParameters().length == 0);
+    } catch (SerializationException e) {
+      e.printStackTrace();
+    }
+
+    // Case 4
+    try {
+      RPCRequest request = RPC.decodeRequest(VALID_ENCODED_REQUEST, A.class);
+      assertEquals(A.class.getMethod("method2", null), request.getMethod());
+      assertTrue(request.getParameters().length == 0);
+    } catch (SerializationException e) {
+      e.printStackTrace();
+    }
+
+    // Case 5
+    try {
+      RPCRequest request = RPC.decodeRequest(INVALID_INTERFACE_REQUEST, B.class);
+    } catch (SecurityException e) {
+      // should get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+    
+    // Case 6
+    try {
+      RPCRequest request = RPC.decodeRequest(INVALID_METHOD_REQUEST, A.class);
+    } catch (SecurityException e) {
+      // should get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+  }
+
+  /**
+   * Tests for method RPC.encodeFailure(Method, Object)
+   * 
+   * Cases:
+   * <ol>
+   * <li>Method == null</li>
+   * <li>Object == null</li>
+   * <li>Method is not specified to throw an exception of the given type</li>
+   * <li>Method is specified to throw an exception of the given type</li>
+   * </ol>
+   * 
+   */
+  public void testEncodeFailure() {
+    // Case 1
+    try {
+      RPC.encodeFailure(null, new Throwable());
+      fail("Expected NullPointerException");
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    Method A_method1 = null;
+    try {
+      A_method1 = A.class.getMethod("method1", null);
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 2
+    try {
+      RPC.encodeFailure(A_method1, null);
+      fail("Expected NullPointerException");
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 3
+    try {
+      RPC.encodeFailure(A.class.getMethod("method1", null),
+          new IllegalArgumentException());
+      fail("Expected UnexpectedException");
+    } catch (UnexpectedException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 4
+    try {
+      String str = RPC.encodeFailure(A.class.getMethod("method1", null),
+          new SerializableException());
+      assertTrue(str.contains("SerializableException"));
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+  }
+
+  /**
+   * Tests for RPC.encodeSuccess(Method, Object)
+   * 
+   * Cases:
+   * <ol>
+   * <li>Method == null</li>
+   * <li>Object == null</li>
+   * <li>Method is not specified to return the given type</li>
+   * <li>Method is specified to return the given type</li>
+   * </ol>
+   */
+  public void testEncodeSuccess() {
+    Method A_method1 = null;
+    Method A_method2 = null;
+    try {
+      A_method1 = A.class.getMethod("method1", null);
+      A_method2 = A.class.getMethod("method2", null);
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 1
+    try {
+      RPC.encodeSuccess(null, new Object());
+      fail("Expected NullPointerException");
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 2
+    try {
+      RPC.encodeSuccess(A_method1, null);
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      fail(e.getMessage());
+    }
+
+    // Case 3
+    try {
+      RPC.encodeSuccess(A_method2, new SerializableException());
+    } catch (IllegalArgumentException e) {
+      int x = 1;
+      // expected to get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+
+    // Case 4
+    try {
+      RPC.encodeSuccess(A_method2, new Integer(1));
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+  }
+
+  /**
+   * Tests for RPC.invoke(Object, Method, Object[])
+   * 
+   * Cases:
+   * <ol>
+   * <li>Method == null</li>
+   * <li>Object does not implement Method</li>
+   * <li>Method parameters do not match given parameters
+   * <li>Method throws exception that it is not specified to
+   * <li>Method throws exception that it is specified to throw
+   * </ol>
+   * @throws NoSuchMethodException 
+   * @throws SecurityException 
+   * 
+   */
+  public void testInvoke() throws SecurityException, NoSuchMethodException {
+    // Case 1
+    try {
+      RPC.invoke(null, null, null);
+    } catch (NullPointerException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+    
+    Method A_method1 = A.class.getMethod("method1", null);
+    
+    // Case 2
+    try {
+      RPC.invoke(new B() {
+        public void method1() {
+        }}, 
+        A_method1, null);
+    } catch (SecurityException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+    
+    // Case 3
+    try {
+      RPC.invoke(new A() {
+        public void method1() throws SerializableException {
+        }
+
+        public int method2() {
+          return 0;
+        }
+
+        public int method3(int val) {
+          return 0;
+        }}, 
+        A_method1, new Integer[] { new Integer(1)});
+    } catch (SecurityException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+
+    // Case 4
+    try {
+      RPC.invoke(new A() {
+        public void method1() throws SerializableException {
+          throw new IllegalArgumentException();
+        }
+
+        public int method2() {
+          return 0;
+        }
+
+        public int method3(int val) {
+          return 0;
+        }}, A_method1, null);
+    } catch (UnexpectedException e) {
+      // expected to get here
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+    
+    // Case 5
+    try {
+      RPC.invoke(new A() {
+        public void method1() throws SerializableException {
+          throw new SerializableException();
+        }
+
+        public int method2() {
+          return 0;
+        }
+
+        public int method3(int val) {
+          return 0;
+        }}, A_method1, null);
+    } catch (Throwable e) {
+      e.printStackTrace();
+      fail(e.getMessage());
+    }
+  }
+}
Index: user/test/com/google/gwt/user/client/rpc/RemoteServiceServletTest.java
===================================================================
--- user/test/com/google/gwt/user/client/rpc/RemoteServiceServletTest.java	(revision 442)
+++ user/test/com/google/gwt/user/client/rpc/RemoteServiceServletTest.java	(working copy)
@@ -50,4 +50,7 @@
       }
     });
   }
+
+  
+ 
 }
Index: user/test/com/google/gwt/user/RPCSuite.java
===================================================================
--- user/test/com/google/gwt/user/RPCSuite.java	(revision 442)
+++ user/test/com/google/gwt/user/RPCSuite.java	(working copy)
@@ -5,6 +5,7 @@
 import com.google.gwt.user.client.rpc.InheritanceTest;
 import com.google.gwt.user.client.rpc.ObjectGraphTest;
 import com.google.gwt.user.client.rpc.ValueTypesTest;
+import com.google.gwt.user.server.rpc.RPCTest;
 
 import junit.framework.Test;
 import junit.framework.TestSuite;
@@ -18,6 +19,7 @@
     suite.addTestSuite(CollectionsTest.class);
     suite.addTestSuite(CustomFieldSerializerTest.class);
     suite.addTestSuite(ObjectGraphTest.class);
+    suite.addTestSuite(RPCTest.class);
     
     return suite;
   }
Index: user/javadoc/com/google/gwt/examples/rpc/server/AdvancedExample.java
===================================================================
--- user/javadoc/com/google/gwt/examples/rpc/server/AdvancedExample.java	(revision 0)
+++ user/javadoc/com/google/gwt/examples/rpc/server/AdvancedExample.java	(revision 0)
@@ -0,0 +1,120 @@
+package com.google.gwt.examples.rpc.server;
+
+import com.google.gwt.user.server.rpc.RPC;
+import com.google.gwt.user.server.rpc.RPCRequest;
+
+import java.io.IOException;
+import java.lang.reflect.InvocationTargetException;
+import java.lang.reflect.Method;
+
+import javax.servlet.ServletException;
+import javax.servlet.http.HttpServlet;
+import javax.servlet.http.HttpServletRequest;
+import javax.servlet.http.HttpServletResponse;
+
+/**
+ * 
+ */
+public class AdvancedExample extends HttpServlet {
+  /**
+   * An example of how you could integrate GWTs RPC functionality without using
+   * the {@link com.google.gwt.user.server.rpc.RemoteServiceServlet}.  Note that
+   * it also shows how the 
+   * 
+   */
+  public void doPost(HttpServletRequest httpRequest,
+      HttpServletResponse httpResponse) throws ServletException, IOException {
+    String payload = readPayloadAsUtf8(httpRequest);
+
+    try {
+      RPCRequest rpcRequest = RPC.decodeRequest(payload);
+
+      Object targetInstance = getInstanceToHandleRequest(httpRequest,
+          rpcRequest);
+
+      Method targetMethod = maybeMapRequestedMethod(targetInstance,
+          rpcRequest.getMethod());
+
+      Object[] targetParameters = maybeMapParameters(rpcRequest.getParameters());
+
+      try {
+        Object result = targetMethod.invoke(targetInstance,
+            targetParameters);
+
+        result = maybeMapResult(rpcRequest.getMethod(), result);
+
+        /*
+         * Encode the object that will be given to the client code's
+         * AsyncCallback::onSuccess(Object) method.
+         */
+        String encodedResult = RPC.encodeSuccess(rpcRequest.getMethod(),
+            result);
+
+        returnSuccess(httpResponse, encodedResult);
+      } catch (IllegalArgumentException e) {
+        throw new SecurityException("Security violation: attempt to "
+            + "invoke method " + targetMethod, e);
+      } catch (IllegalAccessException e) {
+        throw new SecurityException("Security violation: attempt to "
+            + "access inaccessible method " + targetMethod
+            + (targetInstance != null ? " on target " + targetInstance : ""), e);
+      } catch (InvocationTargetException e) {
+        Throwable cause = e.getCause();
+
+        Throwable mappedThrowable = maybeMapThrowable(cause,
+            rpcRequest.getMethod());
+
+        /*
+         * Encode the exception that will be passed back to the client's client
+         * code's AsyncCallback::onFailure(Throwable) method.
+         */
+        String failurePayload = RPC.encodeFailure(rpcRequest.getMethod(),
+            mappedThrowable);
+
+        returnFailure(httpResponse, failurePayload);
+      }
+    } catch (Throwable e) {
+      /*
+       * Return an generic error to the client which will be passed to the 
+       * client code's AsyncCallback::onFailure(Throwable) method.
+       */
+      returnGenericFailure(httpResponse);
+    }
+  }
+
+  private Object getInstanceToHandleRequest(HttpServletRequest httpRequest,
+      RPCRequest rpcRequest) {
+    return null;
+  }
+
+  private Method maybeMapRequestedMethod(Object targetInstance, Method method) {
+    return null;
+  }
+
+  private Object[] maybeMapParameters(Object[] parameters) {
+    return null;
+  }
+
+  private Object maybeMapResult(Method method, Object targetResult) {
+    return null;
+  }
+
+  private Throwable maybeMapThrowable(Throwable cause, Method method) {
+    return null;
+  }
+
+  private String readPayloadAsUtf8(HttpServletRequest httpRequest) {
+    return null;
+  }
+
+  private void returnFailure(HttpServletResponse httpResponse,
+      String failurePayload) {
+  }
+
+  private void returnGenericFailure(HttpServletResponse httpResponse) {
+  }
+
+  private void returnSuccess(HttpServletResponse httpResponse,
+      String encodedResult) {
+  }
+}
Index: user/javadoc/com/google/gwt/examples/rpc/server/CanonicalExample.java
===================================================================
--- user/javadoc/com/google/gwt/examples/rpc/server/CanonicalExample.java	(revision 0)
+++ user/javadoc/com/google/gwt/examples/rpc/server/CanonicalExample.java	(revision 0)
@@ -0,0 +1,19 @@
+package com.google.gwt.examples.rpc.server;
+
+import com.google.gwt.user.client.rpc.SerializationException;
+import com.google.gwt.user.server.rpc.RPC;
+import com.google.gwt.user.server.rpc.RPCRequest;
+import com.google.gwt.user.server.rpc.RemoteServiceServlet;
+
+public class CanonicalExample extends RemoteServiceServlet {
+  /**
+   * Process the RPC request encoded into the payload string and return
+   * a string that encodes either the method return or an exception thrown
+   * by it.
+   */
+  public String processCall(String payload) throws SerializationException {
+    RPCRequest rpcRequest = RPC.decodeRequest(payload, this.getClass());
+    
+    return RPC.invoke(this, rpcRequest.getMethod(), rpcRequest.getParameters());
+  }
+}
Index: user/src/com/google/gwt/user/server/rpc/RemoteServiceServlet.java
===================================================================
--- user/src/com/google/gwt/user/server/rpc/RemoteServiceServlet.java	(revision 442)
+++ user/src/com/google/gwt/user/server/rpc/RemoteServiceServlet.java	(working copy)
@@ -15,22 +15,11 @@
  */
 package com.google.gwt.user.server.rpc;
 
-import com.google.gwt.user.client.rpc.RemoteService;
 import com.google.gwt.user.client.rpc.SerializationException;
-import com.google.gwt.user.server.rpc.impl.ServerSerializableTypeOracle;
-import com.google.gwt.user.server.rpc.impl.ServerSerializableTypeOracleImpl;
-import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamReader;
-import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamWriter;
 
 import java.io.ByteArrayOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
-import java.io.UnsupportedEncodingException;
-import java.lang.reflect.InvocationTargetException;
-import java.lang.reflect.Method;
-import java.util.HashMap;
-import java.util.HashSet;
-import java.util.Set;
 import java.util.zip.GZIPOutputStream;
 
 import javax.servlet.ServletContext;
@@ -40,9 +29,18 @@
 import javax.servlet.http.HttpServletResponse;
 
 /**
- * The servlet base class for your RPC service implementations that
- * automatically deserializes incoming requests from the client and serializes
- * outgoing responses for client/server RPCs.
+ * Base class for servlets that implement GWT RPC services.
+ * 
+ * This class may be subclassed by servlets that directly implement service
+ * interfaces.
+ * 
+ * Or, it may be subclassed by servlets that perform framework-specific routing
+ * of GWT RPC service invocations; such servlets need only override the
+ * processCall method to look up and invoke services as appropriate.
+ * 
+ * Or, the public static utility methods of this servlet (and the RPC class) may
+ * be used in implementing GWT RPC functionality in pre-existing servlet
+ * systems.
  */
 public class RemoteServiceServlet extends HttpServlet {
 
@@ -56,7 +54,6 @@
   private static final String CONTENT_ENCODING_GZIP = "gzip";
   private static final String CONTENT_TYPE_TEXT_PLAIN_UTF8 = "text/plain; charset=utf-8";
   private static final String GENERIC_FAILURE_MSG = "The call failed on the server; see server log for details";
-  private static final HashMap TYPE_NAMES;
 
   /**
    * Controls the compression threshold at and below which no compression will
@@ -64,18 +61,6 @@
    */
   private static final int UNCOMPRESSED_BYTE_SIZE_LIMIT = 256;
 
-  static {
-    TYPE_NAMES = new HashMap();
-    TYPE_NAMES.put("Z", boolean.class);
-    TYPE_NAMES.put("B", byte.class);
-    TYPE_NAMES.put("C", char.class);
-    TYPE_NAMES.put("D", double.class);
-    TYPE_NAMES.put("F", float.class);
-    TYPE_NAMES.put("I", int.class);
-    TYPE_NAMES.put("J", long.class);
-    TYPE_NAMES.put("S", short.class);
-  }
-
   /**
    * Return true if the response object accepts Gzip encoding. This is done by
    * checking that the accept-encoding header specifies gzip as a supported
@@ -107,50 +92,77 @@
   }
 
   /**
-   * Find the invoked method on either the specified interface or any super.
+   * Read the payload as UTF-8 from the request stream.
    */
-  private static Method findInterfaceMethod(Class intf, String methodName,
-      Class[] paramTypes, boolean includeInherited) {
+  private static String readPayloadAsUtf8(HttpServletRequest request)
+      throws IOException, ServletException {
+    int contentLength = request.getContentLength();
+    if (contentLength == -1) {
+      // Content length must be known.
+      throw new ServletException("Content-Length must be specified");
+    }
+
+    String contentType = request.getContentType();
+    boolean contentTypeIsOkay = false;
+    // Content-Type must be specified.
+    if (contentType != null) {
+      // The type must be plain text.
+      if (contentType.startsWith("text/plain")) {
+        // And it must be UTF-8 encoded (or unspecified, in which case we assume
+        // that it's either UTF-8 or ASCII).
+        if (contentType.indexOf("charset=") == -1) {
+          contentTypeIsOkay = true;
+        } else if (contentType.indexOf("charset=utf-8") != -1) {
+          contentTypeIsOkay = true;
+        }
+      }
+    }
+    if (!contentTypeIsOkay) {
+      throw new ServletException(
+          "Content-Type must be 'text/plain' with 'charset=utf-8' (or unspecified charset)");
+    }
+    InputStream in = request.getInputStream();
     try {
-      return intf.getDeclaredMethod(methodName, paramTypes);
-    } catch (NoSuchMethodException e) {
-      if (includeInherited) {
-        Class[] superintfs = intf.getInterfaces();
-        for (int i = 0; i < superintfs.length; i++) {
-          Method method = findInterfaceMethod(superintfs[i], methodName,
-              paramTypes, true);
-          if (method != null) {
-            return method;
-          }
+      byte[] payload = new byte[contentLength];
+      int offset = 0;
+      int len = contentLength;
+      int byteCount;
+      while (offset < contentLength) {
+        byteCount = in.read(payload, offset, len);
+        if (byteCount == -1) {
+          throw new ServletException("Client did not send " + contentLength
+              + " bytes as expected");
         }
+        offset += byteCount;
+        len -= byteCount;
       }
-
-      return null;
+      return new String(payload, "UTF-8");
+    } finally {
+      if (in != null) {
+        in.close();
+      }
     }
   }
 
-  private final Set knownImplementedInterfaces = new HashSet();
-
   private final ThreadLocal perThreadRequest = new ThreadLocal();
 
   private final ThreadLocal perThreadResponse = new ThreadLocal();
 
-  private final ServerSerializableTypeOracle serializableTypeOracle;
-
   /**
    * The default constructor.
    */
   public RemoteServiceServlet() {
-    serializableTypeOracle = new ServerSerializableTypeOracleImpl(
-        getPackagePaths());
   }
 
   /**
-   * This is called internally.
+   * Standard HttpServlet method: handle the POST.
+   * 
+   * This doPost method swallows ALL exceptions, logs them in the
+   * ServletContext, and returns a GENERIC_FAILURE_MSG response with status code
+   * 500.
    */
   public final void doPost(HttpServletRequest request,
       HttpServletResponse response) {
-    Throwable caught;
     try {
       // Store the request & response objects in thread-local storage.
       //
@@ -161,166 +173,63 @@
       //
       String requestPayload = readPayloadAsUtf8(request);
 
+      // Let subclasses see the serialized request.
+      //
+      onBeforeRequestDeserialized(requestPayload);
+
       // Invoke the core dispatching logic, which returns the serialized
       // result.
       //
       String responsePayload = processCall(requestPayload);
 
+      // Let subclasses see the serialized response.
+      //
+      onAfterResponseSerialized(responsePayload);
+
       // Write the response.
       //
       writeResponse(request, response, responsePayload);
       return;
-    } catch (IOException e) {
-      caught = e;
-    } catch (ServletException e) {
-      caught = e;
-    } catch (SerializationException e) {
-      caught = e;
     } catch (Throwable e) {
-      caught = e;
-    }
+      // Allow custom handling of unexpected failures.
+      //
+      onUnexpectedFailure(e);
 
-    respondWithFailure(response, caught);
+      // Send GENERIC_FAILURE_MSG with 500 status.
+      //
+      respondWithFailure(response);
+    } finally {
+      // null the thread-locals to avoid holding request/response
+      //
+      perThreadRequest.set(null);
+      perThreadResponse.set(null);
+    }
   }
 
   /**
+   * Process a call originating from the given request. Uses the RPC.invoke
+   * method to do the actual work.
+   * 
+   * Subclasses may optionally override this method to handle the payload in any
+   * way they desire (by routing the request to a framework component, for
+   * instance). The HTTPServletRequest and HTTPServletResponse can be accessed
+   * via the getThreadLocalRequest() and getThreadLocalResponse() methods.
+   * 
    * This is public so that it can be unit tested easily without HTTP.
+   * 
+   * @param payload the UTF-8 request payload
+   * @throws SerializationException if cannot deserialize the request, or cannot
+   *           serialize the response
+   * @throws SecurityException if there is a problem locating or invoking the
+   *           service method on the destination object
+   * @throws UnexpectedException if the invocation throws a checked exception
+   *           that is not declared in the service method's signature
+   * @throws RuntimeException if the service method throws an unchecked
+   *           exception (the exception will be the one thrown by the service)
    */
   public String processCall(String payload) throws SerializationException {
-
-    // Let subclasses see the serialized request.
-    //
-    onBeforeRequestDeserialized(payload);
-
-    // Create a stream to deserialize the request.
-    //
-    ServerSerializationStreamReader streamReader = new ServerSerializationStreamReader(
-        serializableTypeOracle);
-    streamReader.prepareToRead(payload);
-
-    // Read the service interface
-    //
-    String serviceIntfName = streamReader.readString();
-
-    // TODO(mmendez): need a way to check the type signature of the service intf
-    // Verify that this very servlet implements the specified interface name.
-    //
-    if (!isImplementedRemoteServiceInterface(serviceIntfName)) {
-      // Bad payload, possible hack attempt.
-      //
-      throw new SecurityException(
-          "Blocked attempt to access interface '"
-              + serviceIntfName
-              + "', which is either not implemented by this servlet or which doesn't extend RemoteService; this is either misconfiguration or a hack attempt");
-    }
-
-    // Actually get the service interface, so that we can query its methods.
-    //
-    Class serviceIntf;
-    try {
-      serviceIntf = getClassFromName(serviceIntfName);
-    } catch (ClassNotFoundException e) {
-      throw new SerializationException("Unknown service interface class '"
-          + serviceIntfName + "'", e);
-    }
-
-    // Read the method name.
-    //
-    String methodName = streamReader.readString();
-
-    // Read the number and names of the parameter classes from the stream.
-    // We have to do this so that we can find the correct overload of the
-    // method.
-    //
-    int paramCount = streamReader.readInt();
-    Class[] paramTypes = new Class[paramCount];
-    for (int i = 0; i < paramTypes.length; i++) {
-      String paramClassName = streamReader.readString();
-      try {
-        paramTypes[i] = getClassOrPrimitiveFromName(paramClassName);
-      } catch (ClassNotFoundException e) {
-        throw new SerializationException("Unknown parameter " + i + " type '"
-            + paramClassName + "'", e);
-      }
-    }
-
-    // For security, make sure the method is found in the service interface
-    // and not just one that happens to be defined on this class.
-    //
-    Method serviceIntfMethod = findInterfaceMethod(serviceIntf, methodName,
-        paramTypes, true);
-
-    // If it wasn't found, don't continue.
-    //
-    if (serviceIntfMethod == null) {
-      // Bad payload, possible hack attempt.
-      //
-      throw new SecurityException(
-          "Method '"
-              + methodName
-              + "' (or a particular overload) on interface '"
-              + serviceIntfName
-              + "' was not found, this is either misconfiguration or a hack attempt");
-    }
-
-    // Deserialize the parameters.
-    //
-    Object[] args = new Object[paramCount];
-    for (int i = 0; i < args.length; i++) {
-      args[i] = streamReader.deserializeValue(paramTypes[i]);
-    }
-
-    // Make the call via reflection.
-    //
-    String responsePayload = GENERIC_FAILURE_MSG;
-    ServerSerializationStreamWriter streamWriter = new ServerSerializationStreamWriter(
-        serializableTypeOracle);
-    Throwable caught = null;
-    try {
-      Class returnType = serviceIntfMethod.getReturnType();
-      Object returnVal = serviceIntfMethod.invoke(this, args);
-      responsePayload = createResponse(streamWriter, returnType, returnVal,
-          false);
-    } catch (IllegalArgumentException e) {
-      caught = e;
-    } catch (IllegalAccessException e) {
-      caught = e;
-    } catch (InvocationTargetException e) {
-      // Try to serialize the caught exception if the client is expecting it,
-      // otherwise log the exception server-side.
-      caught = e;
-      Throwable cause = e.getCause();
-      if (cause != null) {
-        // Update the caught exception to the underlying cause
-        caught = cause;
-        // Serialize the exception back to the client if it's a declared
-        // exception
-        if (isExpectedException(serviceIntfMethod, cause)) {
-          Class thrownClass = cause.getClass();
-          responsePayload = createResponse(streamWriter, thrownClass, cause,
-              true);
-          // Don't log the exception on the server
-          caught = null;
-        }
-      }
-    }
-
-    if (caught != null) {
-      responsePayload = GENERIC_FAILURE_MSG;
-      ServletContext servletContext = getServletContext();
-      // servletContext may be null (for example, when unit testing)
-      if (servletContext != null) {
-        // Log the exception server side
-        servletContext.log("Exception while dispatching incoming RPC call",
-            caught);
-      }
-    }
-
-    // Let subclasses see the serialized response.
-    //
-    onAfterResponseSerialized(responsePayload);
-
-    return responsePayload;
+    RPCRequest rpcRequest = RPC.decodeRequest(payload, this.getClass());
+    return RPC.invoke(this, rpcRequest.getMethod(), rpcRequest.getParameters());
   }
 
   /**
@@ -333,6 +242,10 @@
   }
 
   /**
+   * Is the given remote service interface implemented by this endpoint?
+   */
+
+  /**
    * Gets the <code>HttpServletResponse</code> object for the current call. It
    * is stored thread-locally so that simultaneous invocations can have
    * different response objects.
@@ -358,6 +271,28 @@
   }
 
   /**
+   * Override this method to handle the given failure thrown by the processCall
+   * method. By default, the failure is logged.<p/> An "expected failure" is
+   * an exception thrown by a service method that is declared in the signature
+   * of the service method. These exceptions are serialized back to the client,
+   * and are not passed to this method. This method is called only for
+   * exceptions or errors that are not part of the service method's signature,
+   * or that result from SecurityExceptions, SerializationExceptions, or other
+   * failures within the RPC framework. <p/> Note that this method is called
+   * before the GENERIC_FAILURE_MSG response is created. If the desired behavior
+   * is to both send the GENERIC_FAILURE_MSG response AND to rethrow the
+   * exception, then this method should first send the GENERIC_FAILURE_MSG
+   * response itself (using getThreadLocalResponse), and then rethrow the
+   * exception.
+   * 
+   * @param e the exception which was thrown
+   */
+  protected void onUnexpectedFailure(Throwable e) {
+    ServletContext servletContext = getServletContext();
+    servletContext.log("Exception while dispatching incoming RPC call", e);    
+  }
+
+  /**
    * Determines whether the response to a given servlet request should or should
    * not be GZIP compressed. This method is only called in cases where the
    * requestor accepts GZIP encoding.
@@ -379,239 +314,25 @@
   }
 
   /**
-   * @param stream
-   * @param responseType
-   * @param responseObj
-   * @param isException
-   * @return response
-   */
-  private String createResponse(ServerSerializationStreamWriter stream,
-      Class responseType, Object responseObj, boolean isException) {
-    stream.prepareToWrite();
-    if (responseType != void.class) {
-      try {
-        stream.serializeValue(responseObj, responseType);
-      } catch (SerializationException e) {
-        responseObj = e;
-        isException = true;
-      }
-    }
-
-    String bufferStr = (isException ? "{EX}" : "{OK}") + stream.toString();
-    return bufferStr;
-  }
-
-  /**
-   * Returns the {@link Class} instance for the named class.
-   * 
-   * @param name the name of a class or primitive type
-   * @return Class instance for the given type name
-   * @throws ClassNotFoundException if the named type was not found
-   */
-  private Class getClassFromName(String name) throws ClassNotFoundException {
-    return Class.forName(name, false, this.getClass().getClassLoader());
-  }
-
-  /**
-   * Returns the {@link Class} instance for the named class or primitive type.
-   * 
-   * @param name the name of a class or primitive type
-   * @return Class instance for the given type name
-   * @throws ClassNotFoundException if the named type was not found
-   */
-  private Class getClassOrPrimitiveFromName(String name)
-      throws ClassNotFoundException {
-    Object value = TYPE_NAMES.get(name);
-    if (value != null) {
-      return (Class) value;
-    }
-
-    return getClassFromName(name);
-  }
-
-  /**
-   * Obtain the special package-prefixes we use to check for custom serializers
-   * that would like to live in a package that they cannot. For example,
-   * "java.util.ArrayList" is in a sealed package, so instead we use this prefix
-   * to check for a custom serializer in
-   * "com.google.gwt.user.client.rpc.core.java.util.ArrayList". Right now, it's
-   * hard-coded because we don't have a pressing need for this mechanism to be
-   * extensible, but it is imaginable, which is why it's implemented this way.
-   */
-  private String[] getPackagePaths() {
-    return new String[] {"com.google.gwt.user.client.rpc.core"};
-  }
-
-  /**
-   * Returns true if the {@link java.lang.reflect.Method Method} definition on
-   * the service is specified to throw the exception contained in the
-   * InvocationTargetException or false otherwise. NOTE we do not check that the
-   * type is serializable here. We assume that it must be otherwise the
-   * application would never have been allowed to run.
-   * 
-   * @param serviceIntfMethod
-   * @param e
-   * @return is expected exception
-   */
-  private boolean isExpectedException(Method serviceIntfMethod, Throwable cause) {
-    assert (serviceIntfMethod != null);
-    assert (cause != null);
-
-    Class[] exceptionsThrown = serviceIntfMethod.getExceptionTypes();
-    if (exceptionsThrown.length <= 0) {
-      // The method is not specified to throw any exceptions
-      //
-      return false;
-    }
-
-    Class causeType = cause.getClass();
-
-    for (int index = 0; index < exceptionsThrown.length; ++index) {
-      Class exceptionThrown = exceptionsThrown[index];
-      assert (exceptionThrown != null);
-
-      if (exceptionThrown.isAssignableFrom(causeType)) {
-        return true;
-      }
-    }
-
-    return false;
-  }
-
-  /**
-   * Used to determine whether the specified interface name is implemented by
-   * this class without loading the class (for security).
-   */
-  private boolean isImplementedRemoteServiceInterface(String intfName) {
-    synchronized (knownImplementedInterfaces) {
-      // See if it's cached.
-      //
-      if (knownImplementedInterfaces.contains(intfName)) {
-        return true;
-      }
-
-      Class cls = getClass();
-
-      // Unknown, so walk up the class hierarchy to find the first class that
-      // implements the requested interface
-      //
-      while ((cls != null) && !RemoteServiceServlet.class.equals(cls)) {
-        Class[] intfs = cls.getInterfaces();
-        for (int i = 0; i < intfs.length; i++) {
-          Class intf = intfs[i];
-          if (isImplementedRemoteServiceInterfaceRecursive(intfName, intf)) {
-            knownImplementedInterfaces.add(intfName);
-            return true;
-          }
-        }
-
-        // did not find the interface in this class so we look in the
-        // superclass
-        cls = cls.getSuperclass();
-      }
-
-      return false;
-    }
-  }
-
-  /**
-   * Only called from isImplementedInterface().
-   */
-  private boolean isImplementedRemoteServiceInterfaceRecursive(String intfName,
-      Class intfToCheck) {
-    assert (intfToCheck.isInterface());
-
-    if (intfToCheck.getName().equals(intfName)) {
-      // The name is right, but we also verify that it is assignable to
-      // RemoteService.
-      // 
-      if (RemoteService.class.isAssignableFrom(intfToCheck)) {
-        return true;
-      } else {
-        return false;
-      }
-    }
-
-    Class[] intfs = intfToCheck.getInterfaces();
-    for (int i = 0; i < intfs.length; i++) {
-      Class intf = intfs[i];
-      if (isImplementedRemoteServiceInterfaceRecursive(intfName, intf)) {
-        return true;
-      }
-    }
-
-    return false;
-  }
-
-  private String readPayloadAsUtf8(HttpServletRequest request)
-      throws IOException, ServletException {
-    int contentLength = request.getContentLength();
-    if (contentLength == -1) {
-      // Content length must be known.
-      throw new ServletException("Content-Length must be specified");
-    }
-
-    String contentType = request.getContentType();
-    boolean contentTypeIsOkay = false;
-    // Content-Type must be specified.
-    if (contentType != null) {
-      // The type must be plain text.
-      if (contentType.startsWith("text/plain")) {
-        // And it must be UTF-8 encoded (or unspecified, in which case we assume
-        // that it's either UTF-8 or ASCII).
-        if (contentType.indexOf("charset=") == -1) {
-          contentTypeIsOkay = true;
-        } else if (contentType.indexOf("charset=utf-8") != -1) {
-          contentTypeIsOkay = true;
-        }
-      }
-    }
-    if (!contentTypeIsOkay) {
-      throw new ServletException(
-          "Content-Type must be 'text/plain' with 'charset=utf-8' (or unspecified charset)");
-    }
-    InputStream in = request.getInputStream();
-    try {
-      byte[] payload = new byte[contentLength];
-      int offset = 0;
-      int len = contentLength;
-      int byteCount;
-      while (offset < contentLength) {
-        byteCount = in.read(payload, offset, len);
-        if (byteCount == -1) {
-          throw new ServletException("Client did not send " + contentLength
-              + " bytes as expected");
-        }
-        offset += byteCount;
-        len -= byteCount;
-      }
-      return new String(payload, "UTF-8");
-    } finally {
-      if (in != null) {
-        in.close();
-      }
-    }
-  }
-
-  /**
    * Called when the machinery of this class itself has a problem, rather than
    * the invoked third-party method. It writes a simple 500 message back to the
    * client.
    */
-  private void respondWithFailure(HttpServletResponse response, Throwable caught) {
-    ServletContext servletContext = getServletContext();
-    servletContext.log("Exception while dispatching incoming RPC call", caught);
+  private void respondWithFailure(HttpServletResponse response) {
     try {
       response.setContentType("text/plain");
       response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
       response.getWriter().write(GENERIC_FAILURE_MSG);
     } catch (IOException e) {
-      servletContext.log(
-          "sendError() failed while sending the previous failure to the client",
-          caught);
+      getServletContext().log(
+          "respondWithFailure failed while sending the previous failure to the client",
+          e);
     }
   }
 
+  /**
+   * Write the response payload to the response stream.
+   */
   private void writeResponse(HttpServletRequest request,
       HttpServletResponse response, String responsePayload) throws IOException {
 
@@ -633,8 +354,6 @@
         gzipOutputStream.flush();
         response.setHeader(CONTENT_ENCODING, CONTENT_ENCODING_GZIP);
         reply = output.toByteArray();
-      } catch (UnsupportedEncodingException e) {
-        caught = e;
       } catch (IOException e) {
         caught = e;
       } finally {
@@ -645,7 +364,7 @@
           output.close();
         }
       }
-
+      
       if (caught != null) {
         getServletContext().log("Unable to compress response", caught);
         response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
Index: user/src/com/google/gwt/user/server/rpc/UnexpectedException.java
===================================================================
--- user/src/com/google/gwt/user/server/rpc/UnexpectedException.java	(revision 0)
+++ user/src/com/google/gwt/user/server/rpc/UnexpectedException.java	(revision 0)
@@ -0,0 +1,31 @@
+/*
+ * Copyright 2006 Google Inc.
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License"); you may not
+ * use this file except in compliance with the License. You may obtain a copy of
+ * the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+ * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+ * License for the specific language governing permissions and limitations under
+ * the License.
+ */
+package com.google.gwt.user.server.rpc;
+
+/**
+ * The GWT RPC class throws UnexpectedException when a service method, being
+ * invoked by GWT RPC, throws a checked exception that is not in the service
+ * method's signature.  Such exceptions are "unexpected" and the specific
+ * exception will be the cause of the UnexpectedException.
+ */
+public class UnexpectedException extends RuntimeException {
+  /**
+   * Construct an UnexpectedException.
+   */
+  public UnexpectedException (String s, Throwable cause) {
+    super(s, cause);
+  }
+}
Index: user/src/com/google/gwt/user/server/rpc/RPC.java
===================================================================
--- user/src/com/google/gwt/user/server/rpc/RPC.java	(revision 0)
+++ user/src/com/google/gwt/user/server/rpc/RPC.java	(revision 0)
@@ -0,0 +1,542 @@
+/*
+ * Copyright 2006 Google Inc.
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License"); you may not
+ * use this file except in compliance with the License. You may obtain a copy of
+ * the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+ * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+ * License for the specific language governing permissions and limitations under
+ * the License.
+ */
+package com.google.gwt.user.server.rpc;
+
+import com.google.gwt.user.client.rpc.RemoteService;
+import com.google.gwt.user.client.rpc.SerializationException;
+import com.google.gwt.user.server.rpc.impl.ServerSerializableTypeOracle;
+import com.google.gwt.user.server.rpc.impl.ServerSerializableTypeOracleImpl;
+import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamReader;
+import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamWriter;
+
+import java.lang.reflect.Method;
+import java.lang.reflect.InvocationTargetException;
+import java.util.Arrays;
+import java.util.HashMap;
+import java.util.HashSet;
+import java.util.Map;
+import java.util.Set;
+
+/**
+ * A class which enables integration with the RPC system. It exposes methods for
+ * decoding of RPC requests, encoding of RPC responses, and invocation of RPC
+ * calls on service objects. The operations exposed by this class can be reused
+ * by framework implementors such as Spring, and G4jsf to support a wide range
+ * of service invocation policies.
+ * 
+ * The following example demonstrates the canonical way to use this class.
+ * 
+ * {@example com.google.gwt.examples.rpc.server.CanonicalExample#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)}
+ * 
+ * The following example shows a more advanced way of using this class to create
+ * an adapter between GWT RPC entities and POJOs.
+ * 
+ * {@example com.google.gwt.examples.rpc.server.AdvancedExample#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)}
+ */
+public final class RPC {
+
+  /**
+   * Maps primitive wrapper classes to their corresponding primitive class.
+   */
+  private static final Map PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS = new HashMap();
+
+  /**
+   * Oracle used in stream construction. Encapsulates a set of static,
+   * synchronized caches.
+   */
+  private static ServerSerializableTypeOracle serializableTypeOracle;
+
+  /**
+   * Static map of classes to sets of interfaces (e.g. classes). Optimizes
+   * lookup of interfaces for security.
+   */
+  private static Map/* <Class, Set<String> > */serviceToImplementedInterfacesMap;
+
+  private static final HashMap TYPE_NAMES;
+  static {
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Boolean.class, Boolean.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Byte.class, Byte.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Character.class,
+        Character.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Double.class, Double.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Float.class, Float.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Integer.class, Integer.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Long.class, Long.TYPE);
+    PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Short.class, Short.TYPE);
+
+    TYPE_NAMES = new HashMap();
+    TYPE_NAMES.put("Z", boolean.class);
+    TYPE_NAMES.put("B", byte.class);
+    TYPE_NAMES.put("C", char.class);
+    TYPE_NAMES.put("D", double.class);
+    TYPE_NAMES.put("F", float.class);
+    TYPE_NAMES.put("I", int.class);
+    TYPE_NAMES.put("J", long.class);
+    TYPE_NAMES.put("S", short.class);
+
+    serializableTypeOracle = new ServerSerializableTypeOracleImpl(
+        getPackagePaths());
+
+    serviceToImplementedInterfacesMap = new HashMap();
+  }
+
+  /**
+   * Returns a {@link RPCRequest} class that is built by decoding the contents
+   * of the RPC payload.
+   * 
+   * @param payload a string that encodes the request parameters
+   * @return RPCRequest instance which contains all of the RPC request
+   *         parameters
+   * @throws SerializationException if can't decode the payload
+   * @throws SecurityException if the requested interface can't be found, or if
+   *           the requested method can't be found on the requested interface,
+   *           or if the requested interface doesn't extend RemoteService
+   */
+  public static RPCRequest decodeRequest(String payload)
+      throws SerializationException {
+    return decodeRequest(payload, null);
+  }
+
+  /**
+   * Returns an {@link RPCRequest} that is built by decoding the contents of the
+   * RPC payload and optionally verifying that the specified class or interface
+   * implements the
+   * {@link com.google.gwt.user.client.rpc.RemoteService RemoteService}
+   * interface and method that was specified in the request payload.
+   * 
+   * @param payload a string that encodes the request parameters
+   * @param classOrInterface if non-null, this class or interface is checked to
+   *          verify it implements the
+   *          {@link com.google.gwt.user.client.rpc.RemoteService RemoteService}
+   *          interface and method that was requested in the RPC payload; if it
+   *          does not, a SecurityException is thrown. If null, the requested
+   *          interface is looked up from RPC.getClass().getClassLoader()
+   * @return an {@link RPCRequest} instance
+   * 
+   * @throws NullPointerException if the payload is null
+   * @throws IllegalArgumentException if the payload is the empty string
+   * @throws SerializationException if can't decode the payload
+   * @throws SecurityException if the requested interface can't be found
+   *           (subject to the constraintClass lookup rules), or if the
+   *           requested method can't be found on the requested interface, or if
+   *           the requested interface doesn't extend RemoteService
+   */
+  public static RPCRequest decodeRequest(String payload, Class classOrInterface)
+      throws SerializationException {
+    if (payload == null) {
+      throw new NullPointerException("payload cannot be null");
+    }
+
+    if (payload.length() == 0) {
+      throw new IllegalArgumentException("payload cannot be empty");
+    }
+
+    ServerSerializationStreamReader streamReader = new ServerSerializationStreamReader(
+        serializableTypeOracle);
+    streamReader.prepareToRead(payload);
+
+    String serviceIntfName = streamReader.readString();
+
+    if (classOrInterface != null) {
+      if (!implementsInterface(classOrInterface, serviceIntfName)) {
+        // The service does not implement the requested interface
+        throw new SecurityException("Blocked attempt to access interface '"
+            + serviceIntfName + "', which is not implemented by '"
+            + classOrInterface.getName()
+            + "'; this is either misconfiguration or a hack attempt");
+      }
+    }
+
+    Class serviceIntf;
+    try {
+      serviceIntf = getClassFromSerializedName(serviceIntfName);
+      if (!RemoteService.class.isAssignableFrom(serviceIntf)) {
+        // The requested interface is not a RemoteService interface
+        throw new SecurityException(
+            "Blocked attempt to access interface '"
+                + serviceIntfName
+                + "', which doesn't extend RemoteService; ; this is either misconfiguration or a hack attempt");
+      }
+    } catch (ClassNotFoundException e) {
+      throw new SecurityException("Could not locate requested interface "
+          + serviceIntfName + " in default classloader", e);
+    }
+
+    String serviceMethodName = streamReader.readString();
+
+    int paramCount = streamReader.readInt();
+    Class[] parameterTypes = new Class[paramCount];
+
+    for (int i = 0; i < parameterTypes.length; i++) {
+      String paramClassName = streamReader.readString();
+      try {
+        parameterTypes[i] = getClassFromSerializedName(paramClassName);
+      } catch (ClassNotFoundException e) {
+        throw new SerializationException("Unknown parameter " + i + " type '"
+            + paramClassName + "'", e);
+      }
+    }
+
+    Method method = findInterfaceMethod(serviceIntf, serviceMethodName,
+        parameterTypes, true);
+
+    if (method == null) {
+      throw new SecurityException("Could not locate requested method "
+          + serviceMethodName + " with parameters "
+          + Arrays.asList(parameterTypes) + " in interface " + serviceIntf);
+    }
+
+    Object[] parameterValues = new Object[parameterTypes.length];
+    for (int i = 0; i < parameterValues.length; i++) {
+      parameterValues[i] = streamReader.deserializeValue(parameterTypes[i]);
+    }
+
+    return new RPCRequest(method, parameterValues);
+  }
+
+  /**
+   * Returns a string that encodes an exception thrown while invoking a service
+   * method from an RPC call. It is an error to try and return an exception that
+   * is not in the method's list of checked exceptions.
+   * 
+   * @param serviceMethod the method that threw the exception
+   * @param cause the Throwable that was thrown
+   * @return a string that encodes the exception, if the exception was expected
+   * 
+   * @throws NullPointerException if either the serviceMethod or the cause are
+   *           null
+   * @throws SerializationException if the result cannot be serialized
+   * @throws UnexpectedException if the result was an unexpected exception (a
+   *           checked exception not declared in the serviceMethod's signature)
+   */
+  public static String encodeFailure(Method serviceMethod, Throwable cause)
+      throws SerializationException {
+    if (serviceMethod == null) {
+      throw new NullPointerException("serviceMethod cannot be null");
+    }
+
+    if (cause == null) {
+      throw new NullPointerException("cause cannot be null");
+    }
+
+    if (RPC.isExpectedException(serviceMethod, cause)) {
+      return encodeResponse(cause.getClass(), cause, true);
+    } else {
+      throw new UnexpectedException("Service method '" + serviceMethod
+          + "' threw unexpected exception: " + cause.toString(), cause);
+    }
+  }
+
+  /**
+   * Returns a string that encodes the the specified value as the method's
+   * return. The specified return value must be compatible with the service
+   * method's declared return type.
+   * 
+   * @param serviceMethod the method whose result we are encoding
+   * @param returnValue the object that we wish to send back to the client which
+   *          must be a compatible with the method's declared return type
+   * @return a string that encodes the result of calling a service method
+   * 
+   * @throws IllegalArgumentException if the result is not assignable to the
+   *           service method's return type
+   * @throws NullPointerException if the service method is null
+   * @throws SerializationException if the result cannot be serialized
+   */
+  public static String encodeSuccess(Method serviceMethod, Object returnValue)
+      throws SerializationException {
+    if (serviceMethod == null) {
+      throw new NullPointerException("serviceMethod cannot be null");
+    }
+
+    Class methodReturnType = serviceMethod.getReturnType();
+    if (methodReturnType != void.class && returnValue != null) {
+      Class actualReturnType;
+      if (methodReturnType.isPrimitive()) {
+        actualReturnType = getPrimitiveClassFromWrapper(returnValue.getClass());
+      } else {
+        actualReturnType = returnValue.getClass();
+      }
+
+      if (actualReturnType == null
+          || !methodReturnType.isAssignableFrom(actualReturnType)) {
+        throw new IllegalArgumentException("Return type '"
+            + returnValue.getClass().getName()
+            + "' does not match the return type in the method's signature: '"
+            + serviceMethod.toString() + "'");
+      }
+    }
+
+    return encodeResponse(methodReturnType, returnValue, false);
+  }
+
+  /**
+   * Returns a string that encodes the result of calling a service method which
+   * could be the value returned by the method or an exception thrown by it.
+   * 
+   * This method does no security checking; security checking must be done on
+   * the method prior to this invocation.
+   * 
+   * @param target instance on which to call the method
+   * @param serviceMethod the method to invoke the RPC call.
+   * @param parameters parameters to pass as part of the call
+   * @return a string encoding the either the return value or an exception
+   *         thrown by the service method
+   * 
+   * @throws SecurityException if method could not be accessed, or if parameters
+   *           are not type-valid
+   * @throws SerializationException if an object could not be serialized by the
+   *           stream
+   * @throws UnexpectedException if the serviceMethod throws a checked exception
+   *           that is not declared in its signature
+   */
+  public static String invoke(Object target, Method serviceMethod,
+      Object[] parameters) throws SerializationException {
+
+    if (serviceMethod == null) {
+      throw new NullPointerException("serviceMethod cannot be null");
+    }
+
+    String responsePayload;
+    try {
+      Object result = serviceMethod.invoke(target, parameters);
+
+      responsePayload = encodeSuccess(serviceMethod, result);
+    } catch (IllegalAccessException e) {
+      throw new SecurityException("Security violation: attempt to "
+          + "access inaccessible method " + serviceMethod
+          + (target != null ? " on target " + target : ""), e);
+    } catch (IllegalArgumentException e) {
+      throw new SecurityException(
+          "Security violation: attempt to "
+              + "invoke method "
+              + serviceMethod
+              + (target != null ? (" on target " + target) : "")
+              + (parameters != null
+                  ? (" with invalid parameters " + Arrays.asList(parameters))
+                  : ""), e);
+    } catch (InvocationTargetException e) {
+      // Try to serialize the caught exception if the client is expecting it,
+      // otherwise throw UnexpectedException (or RuntimeException, if
+      // unchecked).
+      Throwable cause = e.getCause();
+
+      responsePayload = encodeFailure(serviceMethod, cause);
+    }
+    return responsePayload;
+  }
+
+  /**
+   * Returns a string that encodes the results of an RPC call. Private overload
+   * that takes a flag signaling the preamble of the response payload.
+   * 
+   * @param serviceMethod the method whose return object we are encoding
+   * @param object the object that we wish to send back to the client
+   * @param wasThrown if true, the object being returned was an exception thrown
+   *          by the service method; if false, it was the result of the service
+   *          method's invocation
+   * @return a string that encodes the response from a service method
+   * @throws SerializationException if the object cannot be serialized
+   */
+  private static String encodeResponse(Class responseClass, Object object,
+      boolean wasThrown) throws SerializationException {
+
+    ServerSerializationStreamWriter stream = new ServerSerializationStreamWriter(
+        serializableTypeOracle);
+
+    stream.prepareToWrite();
+    if (responseClass != void.class) {
+      stream.serializeValue(object, responseClass);
+    }
+
+    String bufferStr = (wasThrown ? "{EX}" : "{OK}") + stream.toString();
+    return bufferStr;
+  }
+
+  /**
+   * Find the invoked method on either the specified interface or any super.
+   */
+  private static Method findInterfaceMethod(Class intf, String methodName,
+      Class[] paramTypes, boolean includeInherited) {
+    try {
+      return intf.getDeclaredMethod(methodName, paramTypes);
+    } catch (NoSuchMethodException e) {
+      if (includeInherited) {
+        Class[] superintfs = intf.getInterfaces();
+        for (int i = 0; i < superintfs.length; i++) {
+          Method method = findInterfaceMethod(superintfs[i], methodName,
+              paramTypes, true);
+          if (method != null) {
+            return method;
+          }
+        }
+      }
+
+      return null;
+    }
+  }
+
+  /**
+   * Returns the {@link Class} instance for the named class or primitive type.
+   * 
+   * @param serializedName the serialized name of a class or primitive type
+   * @return Class instance for the given type name
+   * @throws ClassNotFoundException if the named type was not found
+   */
+  private static Class getClassFromSerializedName(String serializedName)
+      throws ClassNotFoundException {
+    Object value = TYPE_NAMES.get(serializedName);
+    if (value != null) {
+      return (Class) value;
+    }
+
+    return Class.forName(serializedName, false, RPC.class.getClassLoader());
+  }
+
+  /**
+   * Obtain the special package-prefixes we use to check for custom serializers
+   * that would like to live in a package that they cannot. For example,
+   * "java.util.ArrayList" is in a sealed package, so instead we use this prefix
+   * to check for a custom serializer in
+   * "com.google.gwt.user.client.rpc.core.java.util.ArrayList". Right now, it's
+   * hard-coded because we don't have a pressing need for this mechanism to be
+   * extensible, but it is imaginable, which is why it's implemented this way.
+   */
+  private static String[] getPackagePaths() {
+    return new String[] {"com.google.gwt.user.client.rpc.core"};
+  }
+
+  /**
+   * Returns the {@link java.lang.Class Class} for a primitive type given its
+   * corresponding wrapper {@link java.lang.Class Class}.
+   * 
+   * @param wrapperClass primitive wrapper class
+   * @return primitive class
+   */
+  private static Class getPrimitiveClassFromWrapper(Class wrapperClass) {
+    return (Class) PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.get(wrapperClass);
+  }
+
+  /**
+   * Used to determine whether the specified interface name is implemented by
+   * the service class. This is done without loading the class (for security).
+   */
+  private static boolean implementsInterface(Class service, String intfName) {
+    synchronized (serviceToImplementedInterfacesMap) {
+      // See if it's cached.
+      //
+      Set/* <String> */interfaceSet = (Set) serviceToImplementedInterfacesMap.get(service);
+      if (interfaceSet != null) {
+        if (interfaceSet.contains(intfName)) {
+          return true;
+        }
+      } else {
+        interfaceSet = new HashSet/* <String> */();
+        serviceToImplementedInterfacesMap.put(service, interfaceSet);
+      }
+
+      if (!service.isInterface()) {
+        while ((service != null) && !RemoteServiceServlet.class.equals(service)) {
+          Class[] intfs = service.getInterfaces();
+          for (int i = 0; i < intfs.length; i++) {
+            Class intf = intfs[i];
+            if (implementsInterfaceRecursive(intf, intfName)) {
+              interfaceSet.add(intfName);
+              return true;
+            }
+          }
+
+          // did not find the interface in this class so we look in the
+          // superclass
+          //
+          service = service.getSuperclass();
+        }
+      } else {
+        if (implementsInterfaceRecursive(service, intfName)) {
+          interfaceSet.add(intfName);
+          return true;
+        }
+      }
+
+      return false;
+    }
+  }
+
+  /**
+   * Only called from implementsInterface().
+   */
+  private static boolean implementsInterfaceRecursive(Class clazz,
+      String intfName) {
+    assert (clazz.isInterface());
+
+    if (clazz.getName().equals(intfName)) {
+      return true;
+    }
+
+    // search implemented interfaces
+    Class[] intfs = clazz.getInterfaces();
+    for (int i = 0; i < intfs.length; i++) {
+      Class intf = intfs[i];
+      if (implementsInterfaceRecursive(intf, intfName)) {
+        return true;
+      }
+    }
+
+    return false;
+  }
+
+  /**
+   * Returns true if the {@link java.lang.reflect.Method Method} definition on
+   * the service is specified to throw the exception contained in the
+   * InvocationTargetException or false otherwise. NOTE we do not check that the
+   * type is serializable here. We assume that it must be otherwise the
+   * application would never have been allowed to run.
+   * 
+   * @param serviceIntfMethod the method from the RPC request
+   * @param cause the exception that the method threw
+   * @return true if the exception's type is in the method's signature
+   */
+  private static boolean isExpectedException(Method serviceIntfMethod,
+      Throwable cause) {
+    assert (serviceIntfMethod != null);
+    assert (cause != null);
+
+    Class[] exceptionsThrown = serviceIntfMethod.getExceptionTypes();
+    if (exceptionsThrown.length <= 0) {
+      // The method is not specified to throw any exceptions
+      //
+      return false;
+    }
+
+    Class causeType = cause.getClass();
+
+    for (int index = 0; index < exceptionsThrown.length; ++index) {
+      Class exceptionThrown = exceptionsThrown[index];
+      assert (exceptionThrown != null);
+
+      if (exceptionThrown.isAssignableFrom(causeType)) {
+        return true;
+      }
+    }
+
+    return false;
+  }
+
+  /**
+   * Static classes have no constructability.
+   */
+  private RPC() {
+  }
+}
Index: user/src/com/google/gwt/user/server/rpc/RPCRequest.java
===================================================================
--- user/src/com/google/gwt/user/server/rpc/RPCRequest.java	(revision 0)
+++ user/src/com/google/gwt/user/server/rpc/RPCRequest.java	(revision 0)
@@ -0,0 +1,57 @@
+/*
+ * Copyright 2006 Google Inc.
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License"); you may not
+ * use this file except in compliance with the License. You may obtain a copy of
+ * the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+ * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+ * License for the specific language governing permissions and limitations under
+ * the License.
+ */
+package com.google.gwt.user.server.rpc;
+
+import java.lang.reflect.Method;
+
+/**
+ * Describes an incoming RPC request in terms of a resolved
+ * java.lang.Method and an array of arguments.
+ */
+public final class RPCRequest {
+
+  /**
+   * The method for this request.
+   */
+  private final Method method;
+
+  /**
+   * The parameters for this request.
+   */
+  private final Object[] parameters;
+
+  /**
+   * Construct an RPCRequest.
+   */
+  public RPCRequest (Method method, Object[] parameters) {
+    this.method = method;
+    this.parameters = parameters;
+  }
+
+  /**
+   * Get the request's method.
+   */
+  public Method getMethod () {
+    return method;
+  }
+
+  /**
+   * Get the request's parameters.
+   */
+  public Object[] getParameters () {
+    return parameters;
+  }
+}
Index: user/src/com/google/gwt/user/server/rpc/package.html
===================================================================
--- user/src/com/google/gwt/user/server/rpc/package.html	(revision 442)
+++ user/src/com/google/gwt/user/server/rpc/package.html	(working copy)
@@ -1,8 +1,28 @@
 <html>
 <body>
 Classes used in server-side implementation of remote procedure calls.
-
-Of primary interest is RemoteServiceServlet, which is the base class of all RPC
-servers.
+<p/>
+The RemoteServiceServlet class provides the most convenient implementation
+of server-side GWT RPC.  This class can be used in two ways:  it can be
+subclassed by servlets that directly implement one or more service
+interfaces, in which case incoming RPC calls will be directed to the
+servlet subclass itself; or it can be overridden to give finer control over
+routing RPC calls within a server framework.  (For more details on the
+latter, see the 
+{@link com.google.gwt.user.server.rpc.RemoteServiceServlet#processCall(String) RemoteServiceServlet.processCall(String)} method.)
+<p/>
+Alternatively, GWT RPC handling can be integrated into an existing framework,
+by using the RPC class to perform GWT invocation decoding and encoding, and
+by using the RemoteServiceServlet static methods to manage the HTTP request
+and response.  RemoteServiceServlet need not be subclassed at all in this
+case, though reading its source is advisable.
+<p/>
+Note that the default RemoteServiceServlet implementation never throws
+exceptions to the servlet container.  All exceptions that escape the the 
+{@link com.google.gwt.user.server.rpc.RemoteServiceServlet#processCall(String) RemoteServiceServlet.processCall(String)}
+method will be caught, logged in the servlet context, and will cause a generic 
+failure message to be sent to the GWT client -- with a 500 status code.  To 
+customize this behavior, override 
+{@link com.google.gwt.user.server.rpc.RemoteServiceServlet#onUnexpectedFailure(java.lang.Throwable) RemoteServiceServlet.onUnexpectedFailure(java.lang.Throwable)}.
 </body>
 </html>
